#include "http_tcp_server.hpp"

#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fstream>

#define LINE_DECOR " ####### "

namespace {
	bool logging=true;
	const int BUFFER_SIZE = 32767;
	
	//Functions
	void log(const std::string &message){
		if(logging==true){
			std::cout << message << std::endl;
		}
    }
	void exitWithError(const std::string &errorMessage){
        log("ERROR: " + errorMessage);
        exit(1);
    }
}

namespace http{
    TcpServer::TcpServer(std::string ip_address, int port) : m_ip_address(ip_address), m_port(port), m_socket(), m_new_socket(), m_incomingMessage(), m_socketAddress(), m_socketAddress_len(sizeof(m_socketAddress)){
        m_socketAddress.sin_family = AF_INET;
        m_socketAddress.sin_port = htons(m_port);
        m_socketAddress.sin_addr.s_addr = inet_addr(m_ip_address.c_str());

        if (startServer() != 0){
            std::ostringstream ss;
            ss << "Failed to start server with PORT: " << ntohs(m_socketAddress.sin_port);
            log(ss.str());
        }
    }

    TcpServer::~TcpServer(){
        closeServer();
    }

    int TcpServer::startServer(){
        m_socket = socket(AF_INET, SOCK_STREAM, 0);
        if (m_socket < 0){
            exitWithError("Cannot create socket");
            return 1;
        }

		int opt = 1;
		if (setsockopt(m_socket, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
			exitWithError("setsockopt(SO_REUSEADDR) failed");
			return 1;
		}
        if (bind(m_socket, (sockaddr *)&m_socketAddress, m_socketAddress_len) < 0){
            exitWithError("Cannot connect socket to address");
            return 1;
        }

        return 0;
    }

    void TcpServer::closeServer(){
        close(m_socket);
        close(m_new_socket);
        return;
    }

	void TcpServer::setContentPath(std::string new_path){
		contentPath=new_path;
		registerRoute("/","index.html");
		log("Content path set to: "+contentPath);
	}
	
	void TcpServer::registerRoute(std::string route, std::string document_path){
		std::string nroute=std::string("/")+route;
		log("Registered new route: "+nroute);
		if(routeMap[nroute]==""){
			routeMap[nroute]=std::string("/")+document_path;
		} else {
			exitWithError("Cannot bind 2 files to the same route!");
		}
	}
	void TcpServer::registerFile(std::string path, std::string name, std::string extension){
		registerRoute(path+std::string("/")+name+std::string(".")+extension,path+std::string("/")+name+std::string(".")+extension);
	}
	
    void TcpServer::startListen(){
        if (listen(m_socket, 20) < 0){
            exitWithError("Socket listen failed");
        }

        std::ostringstream ss;
        ss << "\n" LINE_DECOR "Listening on ADDRESS: " << inet_ntoa(m_socketAddress.sin_addr) << " PORT: " << ntohs(m_socketAddress.sin_port) << LINE_DECOR "\n\n";
        log(ss.str());

        int bytesReceived;

        while (true){
            log(LINE_DECOR "Waiting for a new connection " LINE_DECOR "\n\n\n");
            acceptConnection(m_new_socket);

            char buffer[BUFFER_SIZE] = {0};
            bytesReceived = read(m_new_socket, buffer, BUFFER_SIZE);
            if (bytesReceived < 0)
            {
                exitWithError("Failed to read bytes from client socket connection");
            }

            std::ostringstream ss;
            ss << LINE_DECOR "Received Request from client " LINE_DECOR "\n\n";
            log(ss.str());
			
			std::string request(buffer, bytesReceived);
			std::istringstream requestStream(request);
			std::string method;
			std::string path;
			requestStream >> method >> path;
			log("Requested path: " + path);
			
			if (routeMap.find(path) != routeMap.end()) {
				m_serverMessage = buildResponse(routeMap[path]);
			} else {
				m_serverMessage = buildStatusResponse("404 Not Found", 404);
			}
            sendResponse();

            close(m_new_socket);
			
			std::string cur_input="";
			std::cin >> cur_input;
			if(cur_input=="exit"){
				log("EXIT function called\n");
				break;
			}
        }
    }

    void TcpServer::acceptConnection(int &new_socket){
        new_socket = accept(m_socket, (sockaddr *)&m_socketAddress, &m_socketAddress_len);
        if (new_socket < 0){
            std::ostringstream ss;
            ss << "Server failed to accept incoming connection from ADDRESS: " << inet_ntoa(m_socketAddress.sin_addr) << "; PORT: " << ntohs(m_socketAddress.sin_port);
            exitWithError(ss.str());
        }
    }

    std::string TcpServer::buildResponse(std::string filepath){
		if(contentPath==""){
			exitWithError("Content path is unset!");
		}
		
		std::string filecontent="";
		std::string intext;
		
		std::string fullPath=std::string("./")+contentPath+filepath;

		std::ifstream rfile(fullPath);
		if(!rfile){
			return buildStatusResponse("404 Not Found", 404);
		}
		
		while (getline (rfile, intext)) {
			filecontent+=intext;
		}
		
		std::string htmlFile=filecontent;
        std::ostringstream ss;
        ss << "HTTP/1.1 " << 200 << " " << "OK" << "\n"
		   << "Content-Type: text/html\n"
		   << "Content-Length: " << htmlFile.size() << "\n\n"
		   << htmlFile;

        return ss.str();
    }
	std::string TcpServer::buildStatusResponse(const std::string& body, int statusCode){
		std::ostringstream ss;
		std::string statusText = (statusCode == 200) ? "OK" : "Not Found";
		std::ostringstream html;
		html << "<!DOCTYPE html><html lang=\"en\"><body><h1>" << body << "</h1></body></html>";

		ss << "HTTP/1.1 " << statusCode << " " << statusText << "\n"
		   << "Content-Type: text/html\n"
		   << "Content-Length: " << html.str().size() << "\n\n"
		   << html.str();
		return ss.str();
	}

    void TcpServer::sendResponse(){
        unsigned long bytesSent;

        bytesSent = write(m_new_socket, m_serverMessage.c_str(), m_serverMessage.size());

        if (bytesSent == m_serverMessage.size()){
            log(LINE_DECOR "Server Response sent to client" LINE_DECOR "\n\n");
        }else{
            log("Error sending response to client");
        }
    }

}